use crate::{amount::AmountCents, Amount};
use serde::{Deserialize, Serialize};
use std::{fmt::Display, sync::Arc};

pub type AssetId = u32;

/// An asset
///
/// The asset definition is just a u32 number, which is meaningless for the
/// code, as long as each asset has an unique id.
///
/// The asset struct also has a precision, which is a u8 number (from 0-255)
/// which is used by amount when converting the inner_amount to a human readable
/// format.
#[derive(Clone, Debug, Eq, Hash, PartialEq, Serialize, Deserialize)]
pub struct Asset {
    /// The Asset id
    pub id: AssetId,
    #[serde(
        serialize_with = "crate::serialize_arc_str",
        deserialize_with = "crate::deserialize_arc_str"
    )]
    /// The name of the asset
    pub name: Arc<str>,
    /// The precision
    pub(crate) precision: u8,
}

impl Asset {
    /// Creates a new amount from the asset
    pub fn new_amount(&self, cents: AmountCents) -> Amount {
        Amount::new(self.clone(), cents)
    }
}

impl Display for Asset {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.id)
    }
}
