//! SQLite storage layer for Verax
use crate::{
    amount::AmountCents,
    asset::AssetId,
    changelog::Changelog,
    payment::{self, SpentInfo},
    storage::{Error, Storage},
    transaction::{self, from_db, Type},
    AccountId, Amount, Asset, AssetManager, Payment, PaymentId, Status, TransactionId,
};
use chrono::NaiveDateTime;
use futures::TryStreamExt;
use serde::{de::DeserializeOwned, Serialize};
use sqlx::{sqlite::SqliteRow, Executor, Row, SqliteConnection};
use std::{collections::HashMap, ops::Deref};

mod batch;

pub use batch::Batch;
pub use sqlx::sqlite::{SqliteConnectOptions, SqlitePoolOptions};

/// SQLite storage layer for Verax
pub struct SQLite {
    db: sqlx::SqlitePool,
    asset_manager: AssetManager,
}

impl SQLite {
    /// Creates a new Verax SQLite storage layer. It takes an instance of the asset_manager
    pub fn new(db: sqlx::SqlitePool, asset_manager: AssetManager) -> Self {
        Self { db, asset_manager }
    }

    /// Creates all the tables and indexes that are needed
    pub async fn setup(&self) -> Result<(), sqlx::Error> {
        let mut x = self.db.begin().await?;
        x.execute(
            r#"CREATE TABLE IF NOT EXISTS "payments" (
                "transaction_id" VARCHAR(66) NOT NULL,
                "position_id" INTEGER NOT NULL,
                "asset_id" TEXT NOT NULL,
                "cents" BIG INTEGER NOT NULL,
                "status" INTEGER NOT NULL,
                "to" VARCHAR(71) NOT NULL,
                "spent_by" TEXT,
                "spent_by_status" INTEGER DEFAULT NULL,
                "created_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
                "updated_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
                PRIMARY KEY ("transaction_id", "position_id")
            );
            CREATE INDEX IF NOT EXISTS payments_to ON payments ("to", "asset_id", "spent_by", "status");
            CREATE TABLE IF NOT EXISTS "transactions" (
                "transaction_id" VARCHAR(66) NOT NULL,
                "status" INTEGER NOT NULL,
                "type" INTEGER NOT NULL,
                "reference" TEXT NOT NULL,
                "last_version" BLOB NOT NULL,
                "created_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
                "updated_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
                PRIMARY KEY ("transaction_id")
            );
            CREATE TABLE IF NOT EXISTS "transaction_accounts" (
                "transaction_id" VARCHAR(66) NOT NULL,
                "account_id" VARCHAR(71) NOT NULL,
                "type" INTEGER NOT NULL,
                "created_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
                "updated_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
                PRIMARY KEY("account_id", "transaction_id")
            );
            CREATE INDEX IF NOT EXISTS "type" ON "transaction_accounts" ("account_id", "type", "created_at" DESC);
            CREATE TABLE IF NOT EXISTS "transaction_input_payments" (
                "transaction_id" VARCHAR(66) NOT NULL,
                "payment_transaction_id" VARCHAR(66) NOT NULL,
                "payment_position_id"  INTEGER NOT NULL,
                "created_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
                "updated_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
                PRIMARY KEY ("transaction_id", "payment_transaction_id", "payment_position_id")
            );
            CREATE TABLE IF NOT EXISTS "changelog" (
                "id" BLOB NOT NULL,
                "object_id" BLOB NOT NULL,
                "previous" BLOB DEFAULT NULL,
                "change" BLOB NOT NULL,
                "created_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
                PRIMARY KEY ("id")
            );
            CREATE INDEX IF NOT EXISTS "changelog_object_id" ON "changelog" ("object_id", "created_at");
            CREATE TABLE IF NOT EXISTS "transaction_tags" (
                "transaction_id" VARCHAR(66) NOT NULL,
                "tag" TEXT NOT NULL,
                "type" INTEGER NOT NULL,
                "status" INTEGER NOT NULL,
                "created_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
                "updated_at" DATETIME DEFAULT CURRENT_TIMESTAMP,
                PRIMARY KEY ("transaction_id", "tag")
            );
            CREATE INDEX IF NOT EXISTS "tags_index" ON "transaction_tags" ("tag", "type", "status");
        "#,
        )
        .await
        .expect("valid");
        x.commit().await?;
        Ok(())
    }

    #[inline]
    async fn get_tags(
        &self,
        conn: &mut SqliteConnection,
        transaction: &TransactionId,
    ) -> Result<Vec<String>, Error> {
        Ok(
            sqlx::query(r#"SELECT tag FROM "transaction_tags" WHERE "transaction_id"=?"#)
                .bind(transaction.to_string())
                .fetch_all(&mut *conn)
                .await
                .map_err(|e| Error::Storage(e.to_string()))?
                .into_iter()
                .map(|x| x.try_get::<String, usize>(0))
                .collect::<Result<Vec<_>, _>>()
                .map_err(|e| Error::Storage(e.to_string()))?,
        )
    }

    #[inline]
    async fn get_changelogs_internal<T: DeserializeOwned + Serialize + Send + Sync>(
        &self,
        conn: &mut SqliteConnection,
        object_id: Vec<u8>,
    ) -> Result<Vec<Changelog<T>>, Error> {
        let rows = sqlx::query(
            r#"SELECT  "previous", "change", "created_at", "id" FROM "changelog" WHERE "object_id" = ? ORDER BY "created_at" ASC"#,
        ).bind(&object_id).fetch_all(&mut *conn).await.map_err(|e| Error::Storage(e.to_string()))?;

        let mut results = vec![];

        for row in rows.into_iter() {
            let change: T = bincode::deserialize(
                &row.try_get::<Vec<u8>, usize>(1)
                    .map_err(|_| Error::Storage("Invalid change content".to_string()))?,
            )
            .map_err(|e| Error::Storage(e.to_string()))?;
            let created_at = row
                .try_get::<NaiveDateTime, usize>(2)
                .map_err(|e| Error::InvalidDate(e.to_string()))?
                .and_utc();

            results.push(Changelog::new_from_db(
                row.try_get::<Option<Vec<u8>>, usize>(0)
                    .map_err(|_| Error::Storage("Invalid previous content".to_string()))?,
                object_id.clone(),
                change,
                created_at,
            ));
        }

        Ok(results)
    }

    #[inline]
    async fn sql_row_to_payment(
        &self,
        conn: &mut SqliteConnection,
        row: SqliteRow,
    ) -> Result<Payment, Error> {
        let id = PaymentId {
            transaction: row
                .try_get::<String, usize>(0)
                .map_err(|_| Error::Storage("Invalid payment_id".to_string()))?
                .as_str()
                .try_into()
                .map_err(|_| Error::Storage("Invalid transaction_id length".to_string()))?,
            position: row
                .try_get::<i64, usize>(1)
                .map_err(|_| Error::Storage("Invalid payment_id".to_string()))?
                .try_into()
                .map_err(|_| Error::Storage("Invalid payment_id".to_string()))?,
        };

        let cents = row
            .try_get::<i64, usize>(3)
            .map_err(|_| Error::Storage("Invalid cents".to_string()))?
            .into();

        let changelog = self
            .get_changelogs_internal::<payment::ChangelogEntry>(conn, id.bytes().to_vec())
            .await?;

        let spent_by: Option<TransactionId> = row
            .try_get::<Option<String>, usize>(6)
            .map_err(|_| Error::Storage("Invalid spent_by".to_string()))?
            .map(|s| s.as_str().try_into())
            .transpose()
            .map_err(|_| Error::Storage("Invalid spent_by".to_string()))?;

        let spent_status = row
            .try_get::<Option<u32>, usize>(7)
            .map_err(|_| Error::Storage("Invalid spent_by_status".to_string()))?
            .map(|status| Status::try_from(status))
            .transpose()
            .map_err(|_| Error::Storage("Invalid spent_by_status".to_string()))?;

        if spent_by.is_some() != spent_status.is_some() {
            return Err(Error::Storage(
                "Invalid spent_by and spent_by_status combination".to_string(),
            ));
        }

        Ok(Payment {
            id,
            amount: self
                .asset_manager
                .asset(
                    row.try_get::<String, usize>(2)
                        .map_err(|_| Error::Storage("Invalid asset_id".to_string()))?
                        .parse()
                        .map_err(|_| Error::Storage("Invalid asset_id".to_string()))?,
                )
                .map_err(|e| Error::Storage(e.to_string()))?
                .new_amount(cents),
            to: row
                .try_get::<String, usize>(4)
                .map_err(|_| Error::Storage("Invalid `to`".to_string()))?
                .as_str()
                .try_into()
                .map_err(|_| Error::Storage("Invalid `to`".to_string()))?,
            status: row
                .try_get::<u32, usize>(5)
                .map_err(|_| Error::Storage("Invalid `status`".to_string()))?
                .try_into()
                .map_err(|_| Error::Storage("Invalid status".to_string()))?,
            changelog,
            spent: spent_by.map(|status| SpentInfo {
                by: status,
                status: spent_status.unwrap_or_default(),
            }),
        })
    }
}

#[async_trait::async_trait]
impl Storage for SQLite {
    type Batch<'a> = Batch<'a>
    where
        Self: 'a;

    async fn begin<'a>(&'a self) -> Result<Self::Batch<'a>, Error> {
        self.db
            .begin()
            .await
            .map(|x| Batch::new(x))
            .map_err(|x| Error::Storage(x.to_string()))
    }

    async fn get_payment(&self, id: &PaymentId) -> Result<Payment, Error> {
        let mut conn = self
            .db
            .acquire()
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;

        let row = sqlx::query(
            r#"
            SELECT
                "p"."transaction_id",
                "p"."position_id",
                "p"."asset_id",
                "p"."cents",
                "p"."to",
                "p"."status",
                "p"."spent_by",
                "p"."spent_by_status"
            FROM
                "payments" "p"
            WHERE
                "p"."transaction_id" = ?
                AND "p"."position_id" = ?
            LIMIT 1
            "#,
        )
        .bind(id.transaction.to_string())
        .bind(id.position.to_string())
        .fetch_optional(&mut *conn)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?
        .ok_or(Error::NotFound)?;

        Ok(self.sql_row_to_payment(&mut *conn, row).await?)
    }

    async fn get_balance(&self, account: &AccountId) -> Result<Vec<Amount>, Error> {
        let mut conn = self
            .db
            .acquire()
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;

        let mut result = sqlx::query(
            r#"
            SELECT
                "asset_id",
                SUM("cents") as "cents"
            FROM
                "payments"
            WHERE
                "to" = ? AND "spent_by" IS NULL AND status = ? 
            GROUP BY "asset_id"
            "#,
        )
        .bind(account.to_string())
        .bind::<u32>(Status::Settled.into())
        .fetch(&mut *conn);

        let mut balances = HashMap::<Asset, Amount>::new();

        while let Some(row) = result
            .try_next()
            .await
            .map_err(|e| Error::Storage(e.to_string()))?
        {
            let asset = self
                .asset_manager
                .asset(
                    row.try_get::<String, usize>(0)
                        .map_err(|_| Error::Storage("Invalid asset_id".to_string()))?
                        .parse()
                        .map_err(|_| Error::Storage("Invalid asset_id".to_string()))?,
                )
                .map_err(|e| Error::Storage(e.to_string()))?;

            let cents = row
                .try_get::<i64, usize>(1)
                .map_err(|_| Error::Storage("Invalid cents".to_string()))?
                .into();

            let new_amount = asset.new_amount(cents);

            if let Some(amount) = balances.get_mut(&asset) {
                *amount = amount
                    .checked_add(&new_amount)
                    .ok_or(Error::Storage("amount overflow".to_owned()))?;
            } else {
                balances.insert(asset, new_amount);
            }
        }

        Ok(balances.into_iter().map(|(_, v)| v).collect())
    }

    async fn get_unspent_payments(
        &self,
        account: &AccountId,
        asset: AssetId,
        mut target_amount: AmountCents,
    ) -> Result<Vec<Payment>, Error> {
        let mut conn = self
            .db
            .acquire()
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;
        let results = sqlx::query(
            r#"
            SELECT
                "p"."transaction_id",
                "p"."position_id",
                "p"."asset_id",
                "p"."cents",
                "p"."to",
                "p"."status",
                "p"."spent_by",
                "p"."spent_by_status"
            FROM
                "payments" as "p"
            WHERE
                "p"."to" = ? AND "p"."asset_id" = ? AND status = ? AND "p"."spent_by" IS NULL
            ORDER BY cents ASC
            "#,
        )
        .bind(account.to_string())
        .bind(asset.to_string())
        .bind::<u32>(Status::Settled.into())
        .fetch_all(&mut *conn)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;

        let mut to_return = vec![];
        for row in results.into_iter() {
            let row = self.sql_row_to_payment(&mut *conn, row).await?;
            target_amount -= row.amount.cents();
            to_return.push(row);
            if target_amount <= 0 {
                break;
            }
        }

        if target_amount <= 0 {
            Ok(to_return)
        } else {
            Err(Error::NotEnoughUnspentPayments(target_amount))
        }
    }

    async fn get_transaction(
        &self,
        transaction_id: &TransactionId,
    ) -> Result<from_db::Transaction, Error> {
        let mut conn = self
            .db
            .acquire()
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;

        let transaction_row = sqlx::query(
            r#"
            SELECT
                "t"."status",
                "t"."type",
                "t"."reference",
                "t"."last_version",
                "t"."created_at",
                "t"."updated_at"
            FROM
                "transactions" "t"
            WHERE
                "t"."transaction_id" = ?
            "#,
        )
        .bind(transaction_id.to_string())
        .fetch_optional(&mut *conn)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?
        .ok_or(Error::NotFound)?;

        let results = sqlx::query(
            r#"
            SELECT
                "p"."transaction_id",
                "p"."position_id",
                "p"."asset_id",
                "p"."cents",
                "p"."to",
                "p"."status",
                "p"."spent_by",
                "p"."spent_by_status"
            FROM
                "transaction_input_payments" "tp"
            INNER JOIN
                "payments" "p"
                ON (
                    "tp"."payment_transaction_id" = "p"."transaction_id"
                    AND "tp"."payment_position_id" = "p"."position_id"
                )
            WHERE
                "tp"."transaction_id" = ?
            "#,
        )
        .bind(transaction_id.to_string())
        .fetch_all(&mut *conn)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;

        let mut spend = vec![];
        for row in results.into_iter() {
            spend.push(self.sql_row_to_payment(&mut *conn, row).await?);
        }

        let results = sqlx::query(
            r#"
            SELECT
                "p"."transaction_id",
                "p"."position_id",
                "p"."asset_id",
                "p"."cents",
                "p"."to",
                "p"."status",
                "p"."spent_by",
                "p"."spent_by_status"
            FROM
                "payments" "p"
            WHERE
                "p"."transaction_id" = ?
            "#,
        )
        .bind(transaction_id.to_string())
        .fetch_all(&mut *conn)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?;

        let mut create = vec![];

        for row in results.into_iter() {
            create.push(self.sql_row_to_payment(&mut *conn, row).await?);
        }

        let status = transaction_row
            .try_get::<u32, usize>(0)
            .map_err(|_| Error::Storage("Invalid status".to_string()))?
            .try_into()
            .map_err(|_| Error::Storage("Invalid status".to_string()))?;
        let typ = transaction_row
            .try_get::<u32, usize>(1)
            .map_err(|_| Error::Storage("Invalid type".to_string()))?
            .try_into()
            .map_err(|_| Error::Storage("Invalid type".to_string()))?;

        let reference = transaction_row
            .try_get::<String, usize>(2)
            .map_err(|_| Error::Storage("Invalid reference".to_string()))?;

        let last_change = transaction_row
            .try_get::<Vec<u8>, usize>(3)
            .map_err(|_| Error::Storage("Invalid last_version".to_string()))?;

        let created_at = transaction_row
            .try_get::<NaiveDateTime, usize>(4)
            .map_err(|e| Error::InvalidDate(e.to_string()))?
            .and_utc();

        let updated_at = transaction_row
            .try_get::<NaiveDateTime, usize>(5)
            .map_err(|e| Error::InvalidDate(e.to_string()))?
            .and_utc();

        let changelog = self
            .get_changelogs_internal::<transaction::ChangelogEntry>(
                &mut *conn,
                transaction_id.deref().to_vec(),
            )
            .await?;

        Ok(from_db::Transaction {
            id: transaction_id.clone(),
            spend,
            create,
            typ,
            tags: self.get_tags(&mut *conn, transaction_id).await?,
            last_change,
            changelog,
            status,
            reference,
            created_at,
            updated_at,
        })
    }

    async fn get_transactions(
        &self,
        account: &AccountId,
        types: &[Type],
        _tags: &[String],
    ) -> Result<Vec<from_db::Transaction>, Error> {
        let mut conn = self
            .db
            .acquire()
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;

        let sql = if types.is_empty() {
            r#"SELECT "transaction_id" FROM "transaction_accounts" WHERE "account_id" = ? ORDER BY "created_at" DESC"#.to_owned()
        } else {
            let params = format!("?{}", ", ?".repeat(types.len() - 1));
            format!(
                r#"SELECT "transaction_id" FROM "transaction_accounts" WHERE "account_id" = ? AND "type" IN ({}) ORDER BY "created_at" DESC"#,
                params
            )
        };

        let sql = sqlx::query(&sql).bind(account.to_string());

        let ids = if !types.is_empty() {
            let mut sql = sql;
            for typ in types.into_iter() {
                sql = sql.bind::<u32>(typ.into());
            }
            sql
        } else {
            sql
        }
        .fetch_all(&mut *conn)
        .await
        .map_err(|e| Error::Storage(e.to_string()))?
        .iter()
        .map(|row| {
            let id: Result<TransactionId, _> = row
                .try_get::<String, usize>(0)
                .map_err(|_| Error::Storage("Invalid transaction_id".to_string()))?
                .as_str()
                .try_into();

            id.map_err(|_| Error::Storage("Invalid transaction_id length".to_string()))
        })
        .collect::<Result<Vec<_>, Error>>()?;

        drop(conn);

        let mut transactions = vec![];
        for id in ids.into_iter() {
            transactions.push(self.get_transaction(&id).await?);
        }

        Ok(transactions)
    }

    async fn get_changelogs<T: DeserializeOwned + Serialize + Send + Sync>(
        &self,
        object_id: Vec<u8>,
    ) -> Result<Vec<Changelog<T>>, Error> {
        let mut conn = self
            .db
            .acquire()
            .await
            .map_err(|e| Error::Storage(e.to_string()))?;

        self.get_changelogs_internal(&mut *conn, object_id).await
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::{storage_test_suite, AssetDefinition};
    use sqlx::sqlite::{SqliteConnectOptions, SqlitePoolOptions};
    use std::fs::remove_file;

    async fn get_instance(assets: AssetManager, test: &str) -> SQLite {
        let path = format!("/tmp/{}.db", test);
        let _ = remove_file(&path);
        let settings = path
            .parse::<SqliteConnectOptions>()
            .expect("valid settings")
            .create_if_missing(true);

        let pool = SqlitePoolOptions::new()
            .connect_with(settings)
            .await
            .expect("pool");

        let db = SQLite::new(pool, assets);
        db.setup().await.expect("setup");
        db
    }

    storage_test_suite!();
}
