//! Abstract definition of a virtual machine and their programs to be executed
//! inside the ledger for each transaction
use super::{Error, Storage, Value};
use std::sync::Arc;
use tokio::sync::{Mutex, RwLock};

/// Trait to implement a program and how it will be executed
#[async_trait::async_trait]
pub trait Instance<S: Storage + 'static>: Send + Sync {
    /// Creates a new instance of the program.
    async fn new(state: Arc<S>, code: &str) -> Result<Self, Error>
    where
        Self: Sized;

    /// Returns the code of the program with a given set of arguments. The
    /// return is the execution ID (which is unique per execution) and the
    /// result of the execution.
    async fn run(
        &self,
        execution_id: usize,
        program_name: &str,
        input: Vec<Value>,
    ) -> Result<Value, Error>;
}

pub struct Program<S: Storage + 'static, I: Instance<S>> {
    state: Arc<S>,
    code: String,
    name: String,
    instances: RwLock<Vec<I>>,
    locks: Mutex<Vec<bool>>,
}

impl<S: Storage + 'static, I: Instance<S>> Program<S, I> {
    pub fn new(state: Arc<S>, name: &str, code: &str) -> Result<Self, Error> {
        Ok(Self {
            state,
            code: code.to_owned(),
            name: name.to_owned(),
            instances: RwLock::new(vec![]),
            locks: Mutex::new(vec![]),
        })
    }

    async fn get_available_instance(&self) -> Result<usize, Error> {
        let mut slot_id = None;
        let mut locks = self.locks.lock().await;
        let total_instances = locks.len();
        for (id, busy) in locks.iter_mut().enumerate() {
            if !*busy {
                slot_id = Some(id);
                *busy = true;
                break;
            }
        }
        drop(locks);

        Ok(if let Some(slot_id) = slot_id {
            // we found a free slot, it is already locked, return the id
            slot_id
        } else {
            // no free slot
            if total_instances < 20 {
                // there are few instances, create a new one
                let mut instances = self.instances.write().await;
                instances.push(I::new(self.state.clone(), &self.code).await?);
                self.locks.lock().await.push(true);
                total_instances
            } else {
                // wait until a slot is available
                #[allow(unused_assignments)]
                let mut slot_id = 0;
                loop {
                    let mut slot_id_internal = None;
                    let mut locks = self.locks.lock().await;
                    for (id, busy) in locks.iter_mut().enumerate() {
                        if !*busy {
                            slot_id_internal = Some(id);
                            *busy = true;
                            break;
                        }
                    }
                    drop(locks);
                    if let Some(slot_id_internal) = slot_id_internal {
                        slot_id = slot_id_internal;
                        break;
                    }
                    tokio::time::sleep(tokio::time::Duration::from_micros(1)).await;
                }
                slot_id
            }
        })
    }

    pub async fn run(&self, execution_id: usize, input: Vec<Value>) -> Result<Value, Error> {
        // find a free instance and lock it for execution
        let instance_id = self.get_available_instance().await?;

        // execute program
        let result = self
            .instances
            .read()
            .await
            .get(instance_id)
            .ok_or(Error::InstanceNotFound(self.name.clone(), instance_id))?
            .run(execution_id, &self.name, input)
            .await;

        // remove lock from instance
        let mut locks = self.locks.lock().await;
        locks[instance_id] = false;

        result
    }
}
